/**CFile***********************************************************************

   FileName    [8queen.c]

   PackageName [8quenn]

   Synopsis    [Main program for the 8 Queen Problem with BDDs.]

   Description [Main program for the 8 Queen Problem with BDDs.
     It solve the queen problem using BDDs with CUDD-2.3.1 for a chess board of size n x n.
     n is given as parameter.
     All the solutions are reported.
     ]

   SeeAlso     []

   Author      [Stefano Quer]

   Copyright   [
    Copyright (c) May 19, 2003 by Politecnico di Torino.
    All Rights Reserved. This software is for educational purposes only.
    Permission is given to academic institutions to use, copy, and modify
    this software and its documentation provided that this introductory
    message is not removed, that this software and its documentation is
    used for the institutions' internal research and educational purposes,
    and that no monies are exchanged. No guarantee is expressed or implied
    by the distribution of this code.
    Send bug-reports and/or questions to: quer@polito.it
    ]

******************************************************************************/

#include "util.h"
#include "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define ROW_LEN_MAX 120
#define FAILURE 0
#define SUCCESS 1

/* Print Debugging Information while Running */
#define DEBUG_PRINTING 0

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int fromVet2Mat (int size, int i, int *r, int *c);
static int fromMat2Vet (int size, int r, int c, int *i);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Main program for the 8 queen problem.]

  Description [Main program for the 8 queen problem.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

int
main(
  int  argc,
  char ** argv
  )
{
  int i, r, c, rr, cc, boardSize, boardSize2;
  char name[ROW_LEN_MAX];
  DdManager *ddMgr;
  DdNode *one, **var;
  DdNode *rRel, *cRel, *dRel, *idRel, *finalRel;
  DdNode *tmpBdd, *tmpBdd1, *tmpBdd2;

  /*
   *  Check Command, Print it Out, etc.
   */

  if (argc != 2) {
    fprintf (stdout, "Run Progam as:\n  ./8queen n\n  where n is the chess board size\n");
    fflush (stdout);
    exit (1);
  }

  fprintf (stdout, "# Running Date: ");
  fflush (stdout);
  system ("date");
  if (gethostname (name, ROW_LEN_MAX) == 0) {
    fprintf (stdout, "# HostName: %s\n", name);
  }
  fflush (stdout);

  fprintf (stdout, "# Command: ");
  for (i=0; i<argc; i++) {
    fprintf (stdout, "%s ", argv[i]);
  }
  fprintf (stdout, "\n");
  fprintf (stdout, "# CUDD Version: ");
  Cudd_PrintVersion (stdout);

  /*
   *  Init Board Size and Number of Cells
   */

  boardSize = atoi (argv[1]);
  boardSize2 = boardSize * boardSize;

#if DEBUG_PRINTING
  fprintf (stdout, "Chess Size: %d x %d = %d\n",
    boardSize, boardSize, boardSize2);
#endif

  /*
   *  Init CUDD Manager
   */

  ddMgr = Cudd_Init (0, 0, CUDD_UNIQUE_SLOTS, 32768, 0);
  if (ddMgr == NULL) {
    return (FAILURE);
  }

  Cudd_AutodynEnable(ddMgr, CUDD_REORDER_SIFT);
  Cudd_EnableReorderingReporting (ddMgr);

  one = DD_ONE (ddMgr);

  /*
   *  Create Variable
   */

  var = ALLOC (DdNode *, boardSize2);
  for (r=1; r<=boardSize; r++) {
    for (c=1; c<=boardSize; c++) {
      fromMat2Vet (boardSize, r, c, &i);
#if DEBUG_PRINTING
      {
        int r2, c2;

        fromVet2Mat (boardSize, i, &r2, &c2);
        fprintf (stdout, "%d %d --> %d --> %d %d\n", r, c, i, r2, c2);
      }  
#endif
      var[i] = Cudd_bddNewVar (ddMgr);
      Cudd_Ref (var[i]);
    }
  }

#if DEBUG_PRINTING
  fprintf (stdout, "Number of BDD var = %d\n", Cudd_ReadSize (ddMgr));
#endif

  /*
   *  Main Cycle: Get Relations for Each Cell of the Board
   */

  finalRel = one;
  Cudd_Ref (finalRel);
  for (r=1; r<=boardSize; r++) {
    for (c=1; c<=boardSize; c++) {

      /* Get Relations on Row r */
      rRel = one;
      Cudd_Ref (rRel);
      for (cc=1; cc<=boardSize; cc++) {
        if (cc!=c) {
          fromMat2Vet (boardSize, r, cc, &i);
          tmpBdd = Cudd_bddAnd (ddMgr, rRel, Cudd_Not (var[i]));
          Cudd_Ref (tmpBdd);
          Cudd_RecursiveDeref (ddMgr, rRel);
          rRel = tmpBdd;
        }
      }
      fromMat2Vet (boardSize, r, c, &i);
      tmpBdd = Cudd_bddOr (ddMgr, rRel, Cudd_Not (var[i]));
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, rRel);
      rRel = tmpBdd;

      /* ... And-ing rRel with finalRel */
      tmpBdd = Cudd_bddAnd (ddMgr, finalRel, rRel);
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, finalRel);
      Cudd_RecursiveDeref (ddMgr, rRel);
      finalRel = tmpBdd;

      /* Get Relations on Column c */
      cRel = one;
      Cudd_Ref (cRel);
      for (rr=1; rr<=boardSize; rr++) {
        if (rr!=r) {
          fromMat2Vet (boardSize, rr, c, &i);
          tmpBdd = Cudd_bddAnd (ddMgr, cRel, Cudd_Not (var[i]));
          Cudd_Ref (tmpBdd);
          Cudd_RecursiveDeref (ddMgr, cRel);
          cRel = tmpBdd;
        }
      }
      fromMat2Vet (boardSize, r, c, &i);
      tmpBdd = Cudd_bddOr (ddMgr, cRel, Cudd_Not (var[i]));
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, cRel);
      cRel = tmpBdd;

      /* ... And-ing rRel with finalRel */
      tmpBdd = Cudd_bddAnd (ddMgr, finalRel, cRel);
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, finalRel);
      Cudd_RecursiveDeref (ddMgr, cRel);
      finalRel = tmpBdd;

      /* Get Relations on Diagonal r, c */
      dRel = one;
      Cudd_Ref (dRel);
      for (rr=1; rr<=boardSize; rr++) {
        cc = c + rr - r;
        if (rr!=r && cc>=1 && cc<=boardSize) {
          fromMat2Vet (boardSize, rr, cc, &i);
          tmpBdd = Cudd_bddAnd (ddMgr, dRel, Cudd_Not (var[i]));
          Cudd_Ref (tmpBdd);
          Cudd_RecursiveDeref (ddMgr, dRel);
          dRel = tmpBdd;
        }
      }
      fromMat2Vet (boardSize, r, c, &i);
      tmpBdd = Cudd_bddOr (ddMgr, dRel, Cudd_Not (var[i]));
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, dRel);
      dRel = tmpBdd;

      /* ... And-ing rRel with finalRel */
      tmpBdd = Cudd_bddAnd (ddMgr, finalRel, dRel);
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, finalRel);
      Cudd_RecursiveDeref (ddMgr, dRel);
      finalRel = tmpBdd;

      /* Get Relations on Inverse Diagonal r, c */
      idRel = one;
      Cudd_Ref (idRel);
      for (rr=1; rr<=boardSize; rr++) {
        cc = r + c - rr;
        if (rr!=r && cc>=1 && cc<=boardSize) {
          fromMat2Vet (boardSize, rr, cc, &i);
          tmpBdd = Cudd_bddAnd (ddMgr, idRel, Cudd_Not (var[i]));
          Cudd_Ref (tmpBdd);
          Cudd_RecursiveDeref (ddMgr, idRel);
          idRel = tmpBdd;
        }
      }
      fromMat2Vet (boardSize, r, c, &i);
      tmpBdd = Cudd_bddOr (ddMgr, idRel, Cudd_Not (var[i]));
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, idRel);
      idRel = tmpBdd;
      
      /* ... And-ing rRel with finalRel */
      tmpBdd = Cudd_bddAnd (ddMgr, finalRel, idRel);
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, finalRel);
      Cudd_RecursiveDeref (ddMgr, idRel);
      finalRel = tmpBdd;
    }
  }

  /*
   *  And-ing the "1-queen per Row" Condition
   */

  for (r=1; r<=boardSize; r++) {
    tmpBdd1 = Cudd_Not (one);
    Cudd_Ref (tmpBdd1);
    for (c=1; c<=boardSize; c++) {
      fromMat2Vet (boardSize, r, c, &i);
      tmpBdd2 = Cudd_bddOr (ddMgr, tmpBdd1, var[i]);
      Cudd_Ref (tmpBdd2);
      Cudd_RecursiveDeref (ddMgr, tmpBdd1);
      tmpBdd1 = tmpBdd2;
    }

    /* Build Final Relation */
    tmpBdd2 = Cudd_bddAnd (ddMgr, finalRel, tmpBdd1);
    Cudd_Ref (tmpBdd2);
    Cudd_RecursiveDeref (ddMgr, finalRel);
    Cudd_RecursiveDeref (ddMgr, tmpBdd1);
    finalRel = tmpBdd2;
  }

  /*
   *  Final Printing
   */

  fprintf (stdout, "Final BDD Size: %d\n",
    Cudd_DagSize (finalRel));
  fprintf (stdout, "Number of Solutions: %d\n",
    ((int) Cudd_CountMinterm (ddMgr, finalRel, boardSize2)));
  if (Cudd_CountMinterm (ddMgr, finalRel, boardSize2) >= 1) {
    fprintf (stdout, "Variables Coding (on the chess board): \n");
    for (i=1, r=1; r<=boardSize; r++) {
      for (c=1; c<=boardSize; c++) {
        fprintf (stdout, "%3d", i++);
      }
      fprintf (stdout, "\n");
    }
    fprintf (stdout, "Variables Coding (on the minterm solution): \n");
    for (i=1, r=1; r<=boardSize; r++) {
      for (c=1; c<=boardSize; c++) {
        fprintf (stdout, "%3d", i++);
      }
    }
    fprintf (stdout, "\n");
    fprintf (stdout, "Solutions (with previous encoding): \n");
    Cudd_PrintMinterm (ddMgr, finalRel);
  }

#if DEBUG_PRINTING
  Cudd_PrintDebug (ddMgr, finalRel, 0, 1);
#endif

  Cudd_RecursiveDeref (ddMgr, finalRel);

#if DEBUG_PRINTING
  Cudd_PrintInfo (ddMgr, stdout);
  Cudd_CheckKeys (ddMgr);
  Cudd_CheckZeroRef (ddMgr);
  Cudd_DebugCheck (ddMgr);
#endif

  Cudd_Quit (ddMgr);

  return (SUCCESS);
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [From Linear Size to Bi-Dimensional Array.]

  Description [From Linear Size to Bi-Dimensional Array.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static int
fromVet2Mat (
  int size,
  int i,
  int *r,
  int *c
  )
{
  if (i<1 || i>(size*size)) {
    return (FAILURE);
  }

  *r = (i-1)/size+1;
  *c = (i-1)%size+1;

  return (SUCCESS);
}

/**Function********************************************************************

  Synopsis    [From Bi-Dimensional Array to Linear Size.]

  Description [From Bi-Dimensional Array to Linear Size.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static int
fromMat2Vet (
  int size,
  int r,
  int c,
  int *i
  )
{
  if (r<1 || r>size || c<1 || c>size) {
    return (FAILURE);
  }

  *i = (r-1)*size + c;

  return (SUCCESS);
}

